from wsgiref.headers import Headers as _Headers
from urllib.parse import urlparse
from base64 import b64decode
from typing import AnyStr, Sequence, Optional, Union, Mapping


class Headers(_Headers):
    """ 解决wsgiref.headers键值类型不支持基于str的类型。
    强制转换成str，避免抛出异常。"""
    def __init__(self, headers: Optional[Union[Sequence, Mapping]]=None):
        if headers is None:
            headers = []
        elif isinstance(headers, dict):
            headers = list(headers.items())
        elif isinstance(headers, list):
            headers = list(headers)
        else:
            raise TypeError()

        super().__init__(headers)

    def _convert_string_type(self, value):
        """Convert/check value type."""
        if type(value) is str:
            return value

        return str(value)


class BaseURI:
    def __init__(self, uri: AnyStr, headers):
        self._uri = None
        self._urlparse = None

        self.headers = Headers(headers)
        self.uri = uri

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, value):
        self._uri = value
        self._urlparse = urlparse(value)

    @property
    def urlparse(self):
        return self._urlparse

    @property
    def hostname(self):
        if not self._urlparse:
            return None
        return self._urlparse.hostname

    @property
    def port(self):
        if not self._urlparse:
            return None
        return self._urlparse.port

    @property
    def path(self):
        if not self._urlparse:
            return None
        return self._urlparse.path

    @property
    def scheme(self):
        if not self._urlparse:
            return None
        return self._urlparse.scheme

    protocol = scheme

    @property
    def query(self):
        if not self._urlparse:
            return None
        return self._urlparse.query

    @property
    def netloc(self):
        if not self._urlparse:
            return None
        return self._urlparse.netloc


class SourceURI(BaseURI):
    def __init__(self, id, uri, headers, cookies=None, proxies=None,
                 max_conn=None, range_field=None, name=None, response=None, **kwargs):
        super(SourceURI, self).__init__(uri, headers)
        self.id = id
        self.cookies = cookies
        self.proxies = proxies
        self.max_conn = max_conn

        self.range_field = range_field

        self.name = name
        self._response = URIResponse.loads(response) if response else None
        self.kwargs = kwargs

    def getresponse(self):
        return self._response

    def set_response(self, resp):
        self._response = resp

    def dumps(self):
        kwargs = {
            'id': self.id,
            'uri': self.uri,
            'headers': self.headers.items(),
            'cookies': self.cookies,
            'proxies': self.proxies,
            'max_conn': self.max_conn,
            'range_field': self.range_field,
            'name': self.name,
            'response': self._response and self._response.dumps()
        }
        kwargs.update(self.kwargs)
        return kwargs

    @classmethod
    def loads(cls, dumpy):
        return cls(**dumpy)

    def __repr__(self):
        return f'<SourceURI id={self.id} name={self.name} uri="{self._uri}">'


class URIResponse(BaseURI):
    def __init__(self, uri,
                 headers,
                 code,
                 msg,
                 length,
                 content_type,
                 range,
                 resume_capability,
                 **kwargs):
        """
        Args:
            uri: 响应URI
            headers: 响应头
            code: 响应代码
            msg: 响应消息
            length: 资源总长
            content_type: 资源类型
            range: 资源范围
            resume_capability: 是否支持断点续传
            **kwargs: 额外参数
        """
        super().__init__(uri, headers)
        self.code = code
        self.length = length
        self.range = range
        self.msg = msg
        self.content_type = content_type
        self.resume_capability = resume_capability
        self.kwargs = kwargs

        # extract HTTP headers
        self.etag = self.headers.get('etag')
        self.date = self.headers.get('date')
        self.last_modified = self.headers.get('last-modified')
        self.content_range = self.headers.get('content-range')
        self.content_md5 = self.headers.get('content-md5')
        self.expires = self.headers.get('expires')
        self.md5 = None
        if self.content_md5:
            # RFC1864
            try:
                self.md5 = b64decode(self.content_md5).hex()
            except:
                pass

    def dumps(self):
        kwargs = {
            'uri': self.uri,
            'headers': list(self.headers.items()),
            'code': self.code,
            'length': self.length,
            'range': self.range,
            'content_type': self.content_type,
            'msg': self.msg,
            'resume_capability': self.resume_capability
        }
        kwargs.update(self.kwargs)
        return kwargs

    @classmethod
    def loads(cls, dumpy):
        return cls(**dumpy)

    def __repr__(self):
        return (f"<UriResponse [{self.code} {self.msg}]'{self.uri}' "
                f"range={self.range}, resume_capability={self.resume_capability}>")


class URIs:
    """ 下载源管理器。 """
    def __init__(self):
        self._uris = []

    def __len__(self):
        return len(self._uris)

    def __getitem__(self, item):
        return self._uris.__getitem__(item)

    def __iter__(self):
        return iter(self._uris)

    def put(self, uri,
            headers=None,
            cookies=None,
            proxies=None,
            max_conn=None,
            range_field=None,
            name=None,
            **kwargs):
        """ 添加新的下载源。 """
        put_id = self.__newid()
        if name is None:
            name = str(put_id)
        src_url = SourceURI(put_id, uri, headers, cookies, proxies, max_conn, range_field, name, **kwargs)

        self._uris[put_id] = src_url
        return self._uris[put_id]

    def __newid(self):
        """ 内部用于生成新ID号的函数。 """
        try:
            index = self._uris.index(None)
        except ValueError:
            index = len(self._uris)
            self._uris.append(None)
        return index

    def dumps(self):
        return [uri.dumps() for uri in self._uris]

    @classmethod
    def loads(cls, dumpy):
        uris = cls()
        for uri in dumpy:
            uris._uris.append(SourceURI(**uri))
        return uris

    @classmethod
    def load_from_source_uris(cls, source_uris: Sequence[SourceURI]):
        uris = cls()
        uris.import_uris(source_uris)
        return uris

    def import_uris(self, source_uris: Sequence[SourceURI]):
        for uri in source_uris:
            self.put(uri.uri, uri.headers.items(), uri.cookies, uri.proxies, uri.max_conn,
                     uri.range_field, uri.name, **uri.kwargs)

    def __repr__(self):
        return f'<URIs {self._uris}>'





from wsgiref.headers import Headers
from urllib.parse import urlunparse
from .abstract import AbstractClient
from ..utils import update_range_field
import logging

log = logging.getLogger(__name__)


class BaseHTTPClient(AbstractClient):
    NAME = 'base_http'
    PROTOCOL_SUPPORT = ('http', 'https')
    ASYNC_EXECUTE = None
    TIMEOUT = 10

    def _build_uri_headers(self):
        source_uri = self.source_uri
        pg = self.progress
        uri = source_uri.uri
        headers = Headers(source_uri.headers.items())

        if self.resume_capability is not False:
            range_field = source_uri.range_field
            if range_field is None:
                range_field = {
                    'Range': 'bytes={begin}-{end_with}'
                }

            scheme, netloc, path, params, query, fragment = list(source_uri.urlparse)
            req_range = (pg.begin + pg.walk_length, pg.end)
            query = (query + ''.join(
                [f'{k}={update_range_field(v, req_range)}'
                 for k, v in range_field.items() if k.startswith('&')]
            )).lstrip('&')

            for k, v in range_field.items():
                if not k.startswith('&'):
                    headers.add_header(k, update_range_field(v, req_range))

            # 由于一些浏览器地址栏会直接把空格显示出来而不进行编码，所以这里单独对空格编码。
            uri = urlunparse((scheme, netloc, path, params, query, fragment)).replace(' ', '%20')

        return uri, headers

    def close(self):
        session = self.session
        resp = self.resp
        self.session = None
        self.resp = None
        if resp:
            resp.close()

    async def pause(self):
        self._closed = True

    def validate_token(self, current_resp):
        resp = self.resp
        if resp is None:
            raise ValueError('cannot validate on a unconnected client.')

        source_resp = self.source_uri.getresponse()

        # TODO: 在多下载源的情况下对下载源之间经过资源数据采样校验，通过后作为响应基准
        if source_resp is None:
            raise ValueError('下载源没有经过校验的资源响应基准。')

        validate_name = ['length', 'etag', 'content_md5', 'content_type', 'last_modified']

        if not all([getattr(current_resp, name) == getattr(source_resp, name)
                    for name in validate_name]):
            log.warning([f'{name}: ({getattr(current_resp, name)}) ?= ({getattr(source_resp, name)})'
                         for name in validate_name])
            raise ValueError('connection resource token not match.')
        return True


def content_range_fullsize(content_range):
    """ 从HTTP响应头中的Content-Range中获取文件总长。"""
    if content_range is None:
        return None
    return int(content_range.rsplit('/', 1)[-1])


def content_type_mimetype(content_type_header):
    """ 从HTTP响应头中的Content-Type中获取文件mimetype类型。"""
    if content_type_header is None:
        return None
    return content_type_header.split(';', 1)[0] or None





class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return f'<{self.__class__.__name__} args={self.args} kwargs={self.kwargs}>'


class GatherableError(Error):
    def __init__(self, exception, exc_info, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exception = exception
        self.exc_info = exc_info

    def __repr__(self):
        return f'<{self.__class__.__name__} exc="{self.exception}">'


class ClientError(GatherableError):
    pass


class HandlerError(GatherableError):
    pass


class UriError(Error):
    pass


class TimeoutError(UriError):
    pass


class FatalError(UriError):
    pass


class MaxRetriesExceeded(ClientError):
    pass



import asyncio
from collections import defaultdict
from contextvars import ContextVar
from contextlib import asynccontextmanager, contextmanager
from concurrent.futures.thread import ThreadPoolExecutor
from nbdler.error import HandlerError, ClientError
from functools import partial
from copy import copy
from operator import attrgetter
import threading
from nbdler.utils import UsageInfo
from traceback import format_exc
import logging
import weakref
import json

log = logging.getLogger(__name__)

block_context = ContextVar('block context')


def _lookup_block():
    """ 查找上下文的下载块。"""
    return block_context.get()


def await_coroutine_threadsafe(coro, timeout=None):
    """ 线程安全等待协程结束。
    Args:
        coro: 协程
        timeout: 等待超时事件

    Returns:
        返回协程的执行结果，或抛出异常。
    """
    fut = asyncio.run_coroutine_threadsafe(coro, h.loop)
    return fut.result(timeout)


class Handlers(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ready = None

    def __getattr__(self, item):
        return self[item]

    async def prepare(self):
        """ Handler启动预处理，通常预启动，做初始化工作。启动标志在该方法设置。"""
        self._ready = asyncio.Event(loop=asyncio.get_running_loop())
        result = await asyncio.gather(*[handler.prepare() for handler in self.values()])
        self._ready.set()
        return result

    async def start(self):
        """ 此方法用于启动Handler的异步工作Handler.run()方法。"""
        result = await asyncio.gather(*[handler.start() for handler in self.values()])
        self._ready = None
        return result

    async def close(self):
        await self._wait_for_ready()
        return await asyncio.gather(*[handler.close() for handler in self.values()])

    async def _wait_for_ready(self):
        """ 等待Handler准备就绪。"""
        ready = self._ready
        if ready is not None:
            await ready.wait()

    async def pause(self):
        await self._wait_for_ready()
        return await asyncio.gather(*[handler.pause() for handler in self.values()])

    async def join(self):
        await self._wait_for_ready()
        return await asyncio.gather(*[handler.join() for handler in self.values()])


class _HandlerReference(threading.local):
    # handlers 异步上下文。
    __context__ = ContextVar('handlers dict')

    def __init__(self):
        # handlers 线程上下文事件循环
        self._loop = None

    @property
    def loop(self):
        return self._loop

    @property
    def owner(self):
        return self.__context__.get()

    @contextmanager
    def enter(self, handlers, loop=None):
        assert self._loop or loop
        if loop:
            if not isinstance(loop, weakref.ProxyType):
                loop = weakref.proxy(loop)
            self._loop = loop

        if not isinstance(handlers, weakref.ProxyType):
            handlers = weakref.proxy(handlers)

        token = self.__context__.set(handlers)

        yield self
        self.__context__.reset(token)

    def __getattr__(self, item):
        return self.__context__.get()[item]

    def __iter__(self):
        return iter(self.__context__.get().values())

    iter_all = __iter__


h = _HandlerReference()


class Handler:
    name = None

    parent = None
    _future = None

    def add_parent(self, parent):
        self.parent = parent

    async def prepare(self, *args, **kwargs):
        pass

    async def start(self):
        assert not self._future or self._future.done()
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        self._future = future
        try:
            result = await self.run()
        finally:
            future.set_result(None)

    async def run(self, *args, **kwargs):
        raise NotImplementedError

    async def pause(self, *args, **kwargs):
        raise NotImplementedError

    async def close(self, *args, **kwargs):
        raise NotImplementedError

    def __repr__(self):
        return f'<Handler name={self.name}>'

    async def join(self):
        return await self._future

    def info_getter(self):
        return None


# TODO: 在多下载源的情况下对下载源之间经过资源数据采样校验，通过后作为响应基准
class SampleValidate(Handler):
    name = 'uri_validate'


class GatherException(Handler):
    """ 下载异常状态集合。

    负责工作：
        1. 收集下载过程中发生的异常
        2. 提供对异常的外部推送
    """
    name = 'exception'

    def __init__(self):
        # 实现异常获取的线程安全
        self._exceptions = defaultdict(list)
        self._cond = threading.Condition(threading.Lock())

        # async 异步异常获取回调
        self._waiter_callbacks = set()

        self._stopped = False

    def handler_error(self, exception):
        """ 推送handler异常

        Args:
            exception:  发生的异常对象
        """
        with self._cond:
            self._exceptions[HandlerError].append(
                HandlerError(exception, format_exc()))
            # 释放线程锁
            self._cond.notify_all()

            # 释放异步锁
            for waiter in self._waiter_callbacks:
                waiter()

    def client_error(self, exception):
        """ 推送client客户端异常

        Args:
            exception:  发生的异常对象
        """
        with self._cond:
            self._exceptions[ClientError].append(
                ClientError(exception, format_exc()))
            # 释放线程锁
            self._cond.notify_all()
            # 释放异步锁
            for waiter in self._waiter_callbacks:
                waiter()

    def _fetch_exceptions(self, exception_type=None):
        if exception_type is None:
            exceptions = []
            for v in self._exceptions.values():
                exceptions.extend(v)
        else:
            exceptions = list(self._exceptions[exception_type])

        return exceptions

    def acquire_threadsafe(self, exception_type=None, *, just_new_exception=True):
        """ 线程安全获取异常

        以生成器的形式获取内部发生的异常，当下载任务暂停或者完成后将中断生成器的迭代。

        Args:
            exception_type: 指定异常类型，可选ClientError、HandlerError。默认None则获取所有异常。
            just_new_exception: 是否忽略当前时间前的旧异常，仅返回之后的新异常。

        Yields:
            内部出现的client或handler异常对象。
        """
        old_exc_list = []
        if just_new_exception:
            old_exc_list = self._fetch_exceptions()
        while True:
            with self._cond:
                if self._stopped:
                    break

                # 在上一次异常推送过程中是否有新的异常被忽略
                # 如果有忽略的异常就不需要等待，先处理被忽略的异常
                before_new_exc = self._fetch_exceptions()
                before_new_diff = sorted(
                    set(before_new_exc).difference(old_exc_list),
                    key=before_new_exc.index)

                if not before_new_diff:
                    self._cond.wait()

            if not before_new_diff:
                new_exc_list = self._fetch_exceptions(exception_type)
                new_exc_set = set(new_exc_list).difference(old_exc_list)
            else:
                new_exc_list = before_new_exc
                new_exc_set = before_new_diff

            if not new_exc_set:
                continue

            for exc in sorted(new_exc_set, key=new_exc_list.index):
                yield exc
            old_exc_list = new_exc_list

    async def acquire(self, exception_type=None, *, just_new_exception=True):
        """ 异步获取异常

        acquire的异步化方法，具体作用参看acquire()方法。

        Args:
            exception_type: 指定异常类型，可选ClientError、HandlerError。默认None则获取所有异常。
            just_new_exception: 是否忽略当前时间前的旧异常，仅返回之后的新异常。

        Yields:
            内部出现的client或handler异常对象。
        """
        def release_waiter():
            nonlocal cond, loop

            async def _release():
                async with cond:
                    cond.notify_all()
            asyncio.run_coroutine_threadsafe(_release(), loop=loop)

        loop = asyncio.get_running_loop()
        cond = asyncio.Condition(asyncio.Lock())
        self._waiter_callbacks.add(release_waiter)

        old_exc_list = []
        if not just_new_exception:
            old_exc_list = self._fetch_exceptions(exception_type)
        while True:
            with self._cond:
                if self._stopped:
                    break
                before_new_exc = self._fetch_exceptions()
                before_new_diff = sorted(
                    set(before_new_exc).difference(old_exc_list),
                    key=before_new_exc.index)

            if not before_new_diff:
                async with cond:
                    await cond.wait()

            if not before_new_diff:
                new_exc_list = self._fetch_exceptions(exception_type)
                new_exc_set = set(new_exc_list).difference(old_exc_list)
            else:
                new_exc_list = before_new_exc
                new_exc_set = before_new_diff

            if not new_exc_set:
                continue

            for exc in sorted(new_exc_set, key=new_exc_list.index):
                yield exc
            old_exc_list = new_exc_list

    async def run(self):
        self._stopped = False
        self._exceptions.clear()

    async def close(self):
        pass

    async def pause(self):
        self._stopped = True
        with self._cond:
            self._cond.notify_all()

        for waiter in self._waiter_callbacks:
            waiter()

    def __repr__(self):
        count = {k: len(v) for k, v in self._exceptions.items()}
        return f'<GatherException {count} future={self._future}>'


class URIStatus:
    def __init__(self, uri):
        self.source_uri = uri
        self._used = 0
        self._success = 0
        self._timeout = 0
        self._fatal = 0

        self._logs = []
        self._users = {}

        self._conn_delay_moving_avg = [0 for _ in range(8)]
        self._conn_delay = float('inf')

    def log(self, resp):
        self._logs.append(resp)

    def _response_delay(self, time_s):
        moving_avg = self._conn_delay_moving_avg
        moving_avg.append(time_s)
        moving_avg.pop(0)
        failure_count = moving_avg.count(float('inf'))
        if failure_count >= 5:
            self._conn_delay = float('inf')
        else:
            self._conn_delay = sum([delay for delay in moving_avg
                                    if delay != float('inf')]) / (8 - failure_count)

    def use(self, block):
        self._used += 1
        self._users[block] = UsageInfo(lambda: block.progress.walk_length)

    def timeout(self, block, resp):
        self._timeout += 1
        self.log(f'{block} {resp}')

    def success(self, block, resp):
        self._success += 1
        self.log(f'{block} {resp}')
        # TODO: 在多下载源的情况下对下载源之间经过资源数据采样校验，通过后作为响应基准
        if self.source_uri.getresponse() is None:
            self.source_uri.set_response(resp)
        self._response_delay(self._users[block].timelength())

    def fatal(self, block, resp):
        self._fatal += 1
        self.log(f'{block} {resp}')

    def disuse(self, block):
        self._used -= 1
        del self._users[block]

    def is_available(self):
        """ 返回当前下载源是否超过有效使用次数。 """
        return self.source_uri.max_conn is None or self.source_uri.max_conn > self._used

    @property
    def users(self):
        return self._users

    def get_copy(self):
        """ 返回URI下载源对象的副本。"""
        return copy(self.source_uri)

    def transfer_rate(self):
        """ 返回下载源的传输速率。"""
        return sum([user.rate for user in self._users.values()])

    def average_speed(self):
        """ 返回当下载源的平均连接的传输速率。"""
        users = [user.rate for user in self._users.values()]
        return sum(users) / len(users)

    def refresh(self):
        """ 刷新当前下载源的状态使用信息。"""
        for user in self._users.values():
            user.refresh()

    def __repr__(self):
        return (f'<URIStatus {self.transfer_rate() / 1024} kb/s [{self._conn_delay * 1000} ms]'
                f'(used={self._used}, success={self._success}, timeout={self._timeout}, fatal={self._fatal})>')

    def info(self):
        return {
            'transfer_rate': self.transfer_rate(),
            'used': self._used,
            'success': self._success,
            'timeout': self._timeout,
            'fatal': self._fatal,
            'connection_delay': self._conn_delay
        }


class URIStatusManager(Handler):
    """ URI状态管理器使用URIStatus对象管理URI下载源。

    负责工作：
        1. 管理和调配URI下载源
        2. 监控URI下载源工作状态
    """

    name = 'uri_mgr'

    def __init__(self):
        self._uri_status = {}
        self._cond = None
        self._stopped = False

    async def prepare(self):
        self._cond = asyncio.Condition()
        for uri in self.parent.uris:
            self._uri_status.setdefault(uri.id, URIStatus(uri))

    async def get_uri(self):
        """  返回URI状态对象供客户端使用。

        以下载源的使用次数为主，尽可能的覆盖所有的下载源，之后根据单连接的下载源传输速度快慢
        分配下载源。

        Returns:
            被分配的URIStatus对象。
        """
        avl_uris = self._find_avl_uris()
        while not avl_uris:
            await self._cond.wait()
            avl_uris = self._find_avl_uris()

        uri = avl_uris[0]
        if uri._used > 0:
            uri = sorted(avl_uris, key=lambda u: u.average_speed(), reverse=True)[0]
        return uri

    def _find_avl_uris(self):
        more_used = sorted(self._uri_status.values(), key=attrgetter('_used'))
        return list(filter(lambda u: u.is_available(), more_used))

    def success(self, resp):
        block = _lookup_block()
        self._uri_status[block.current_uri().id].success(block, resp)

    def timeout(self, resp):
        block = _lookup_block()
        self._uri_status[block.current_uri().id].timeout(block, resp)

    def fatal(self, resp):
        block = _lookup_block()
        self._uri_status[block.current_uri().id].fatal(block, resp)

    async def run(self):
        self._stopped = False
        async_sleep = asyncio.sleep

        uri_status = self._uri_status
        while True:
            await async_sleep(1)
            if self._stopped:
                break

            for status in uri_status.values():
                status.refresh()

        self._cond = None

    async def pause(self):
        self._stopped = True

    async def close(self):
        pass

    def __repr__(self):
        return f'<URIStatusManager {self._uri_status} future={self._future}>'

    def info_getter(self):
        return {k: v.info() for k, v in self._uri_status.items()}


class ClientWorker(Handler):
    """ （主处理器）异步客户端调配工作器。

    负责工作：
        1. 客户端会话管理
        2. 下载块工作调配
        3. 工作进度检测
    """
    name = 'client_worker'

    def __init__(self):
        self._block_queue = None
        self._working_blocks = set()
        self._client_session = {}
        self._stopped = False
        self._executors = None
        self._tasks = set()

    async def prepare(self):
        self._stopped = False

        self._block_queue = asyncio.Queue()
        self._executors = ThreadPoolExecutor(
            max_workers=self.parent.config.max_concurrent,
            thread_name_prefix=self.parent.file.name
        )

    async def run(self):
        def goto_work(blo):
            """ 后台执行下载块。 """
            def cb(fut):
                # 回调移除工作下载块，并交由下载块检测
                self._block_queue.put_nowait(blo)
                self._working_blocks.remove(blo)

            task = asyncio.run_coroutine_threadsafe(
                self._worker(blo), loop)
            self._working_blocks.add(blo)
            task.add_done_callback(cb)
            return task

        loop = asyncio.get_running_loop()
        config = self.parent.config
        block_group = self.parent.block_grp

        # 准备未完成的下载块
        unfinished_blocks = block_group.unfinished_blocks()
        # 提交下载块到工作区
        while unfinished_blocks:
            block = unfinished_blocks.pop(0)
            await self.submit(block)

        # 对下载块做出处理决策
        work_queue = self._block_queue
        resume_capability = config.resume_capability
        while True:
            block = await work_queue.get()
            work_queue.task_done()
            if block is None:
                break
            if block.unused_length():
                # 重试下载块
                goto_work(block)
            else:
                # 检查任务是否完成
                if block_group.is_walk_finished():
                    missing = block_group.integrity_check()
                    if missing:
                        h.exception.handler_error(RuntimeError(f'Missing Blocks: {missing}'))
                    if unfinished_blocks:
                        await work_queue.put(unfinished_blocks.pop(0))
                        continue
                    break
                # 已完成其中一下载块后允许对未完成下载块进行切片补充并发量
                if resume_capability:
                    if len(block_group.unfinished_blocks()) < config.max_concurrent:
                        if unfinished_blocks:
                            goto_work(unfinished_blocks.pop(0))
                        else:
                            h.slicer.request()

        # 任务完成或暂停，清除冗余队列信息
        while not work_queue.empty():
            await work_queue.get()
            work_queue.task_done()

        # 待所有下载块退出
        while self._working_blocks:
            await work_queue.get()
            work_queue.task_done()

        self._executors.shutdown(False)
        # 非阻塞执行关闭所有handler
        self.parent.pause(0)

    async def submit(self, block):
        """ 提交下载块到工作区。
        Args:
            block: 下载块Block对象。
        """
        if self._stopped:
            return False
        await self._block_queue.put(block)

    async def _worker(self, block):
        """ 客户端工作worker。
        Args:
            block: 下载块Block对象。
        """
        def run_client_threadsafe():
            nonlocal cli, loop, handlers_ref, block
            token = block_context.set(block)
            with h.enter(handlers_ref, loop):
                try:
                    return cli.run()
                except BaseException as e:
                    h.exception.client_error(e)
                finally:
                    block_context.reset(token)

        if self._stopped:
            return
        handlers_ref = h.owner
        loop = asyncio.get_running_loop()
        config = self.parent.config

        # 准备下载源
        uri = await h.uri_mgr.get_uri()

        source_uri = uri.get_copy()
        resume_capability = config.resume_capability
        client_policy = config.client_policy

        solution = client_policy.get_solution(source_uri.protocol)

        # 准备客户端会话
        session = self._client_session.get(solution, None)
        if session is None:
            session = solution.get_session()
            self._client_session[solution] = session

        # 准备客户端处理器
        client = solution.get_client(
            session, source_uri, block.progress, resume_capability)

        # 为下载块准备客户端进行下载
        async with block.request(client) as cli:
            uri.use(block)
            try:
                if solution.is_async():
                    fut = cli.run()
                else:
                    fut = loop.run_in_executor(self._executors, run_client_threadsafe)
                result = await fut
            except BaseException as err:
                h.exception.client_error(err)
            uri.disuse(block)
        return result

    async def close(self):
        async def close_sess(sess):
            """ 关闭客户端会话。"""
            coro_or_result = sess.close()
            if asyncio.iscoroutine(coro_or_result):
                await coro_or_result
        await asyncio.gather(*[close_sess(session) for session in self._client_session.values()])
        self._client_session.clear()

    async def pause(self):
        async def pause_cli(blo):
            """ 安全暂停关闭客户端。"""
            while True:
                if blo not in self._working_blocks:
                    # 若下载块已退出，跳过客户端暂停
                    return
                if blo.client is None:
                    # 等待客户端进入，同时关闭
                    await asyncio.sleep(0)
                else:
                    break
            await blo.client.pause()

        if not self._stopped:
            self._stopped = True
            await asyncio.gather(*[pause_cli(block) for block in self._working_blocks])
            await self._block_queue.put(None)

    def __repr__(self):
        return f'<ClientWorker {self._working_blocks} future={self._future}>'

    def info_getter(self):
        return {
            'actives': set(self._working_blocks)
        }


class BlockSlicer(Handler):
    """ 下载块切片器。

    负责工作：
        1. 下载块切片请求和响应
    """
    name = 'slicer'

    def __init__(self):
        self._waiters = set()
        self._lock = threading.Lock()

    async def divide_into(self, n):
        """ 下载块切片器分成n份。

        该方法不建议在在下载块工作过程中进行调用，
        否则可能会出现传输数据冗余的问题。

        Args:
            n: 分成n份。
        """
        for i in range(n):
            self.request()
            while self._waiters:
                block = self._waiters.pop()
                self._slice(block)

    def _slice(self, source_block):
        req_range = source_block.half_unused()
        if req_range:
            result = source_block.slice(req_range)
            if result:
                block = self.parent.block_grp.insert(result)
                return block

        return None

    async def response(self):
        """ 客户端用于响应切片器是否需要对其进行切片。

        若切片器希望对当前下载块进行切片，调用该方法允许切片器安全的对当前下载块进行切片。
        安全的前提是该方法在不影响下载区间的地方调用。
        若当前下载块未在期望切片队列则直接跳过。
        """
        if self._waiters:
            with self._lock:
                source_block = _lookup_block()
                if source_block not in self._waiters:
                    return
                self._waiters.remove(source_block)
                resp = self._slice(source_block)
            if resp is not None:
                await h.client_worker.submit(resp)

    def response_threadsafe(self):
        with self._lock:
            if not self._waiters and _lookup_block() not in self._waiters:
                return False
        await_coroutine_threadsafe(self.response())

    def request(self):
        """ 请求一次下载块切片。

        使用最大剩余block大小策略来选择被切块对象，该方法并未对下载块进行切片，
        需要客户端配合response()方法来响应切片请求。
        """
        len_waiting = len(self._waiters)
        blocks = sorted(self.parent.block_grp.unfinished_blocks(), key=lambda i: i.unused_length(), reverse=True)
        self._waiters = set(blocks[:len_waiting + 1])
        return len(self._waiters) == len_waiting + 1

    async def prepare(self):
        # 下载块切片以保证下载块的最大并发量。
        config = self.parent.config
        if config.resume_capability:
            blocks_len = len(self.parent.block_grp.unfinished_blocks())
            if blocks_len < config.max_concurrent:
                await self.divide_into(config.max_concurrent - blocks_len)

    async def run(self):
        pass

    async def close(self):
        pass

    async def pause(self):
        self._waiters.clear()

    def __repr__(self):
        return f'<BlockSlicer {self._waiters} future={self._future}>'

    def info_getter(self):
        return {
            'waiters': set(self._waiters)
        }


class SpeedAdjuster(Handler):
    """ 速度调节器。

    负责工作：
        1. 最大速度限制
        2. 实时速度信息刷新
    """
    name = 'speed_adjuster'

    def __init__(self):
        self._opened = False
        self._stopped = True
        self._thread_cond = threading.Condition(threading.RLock())
        self._sema_value = 0
        self._async_cond = None

    async def _release_all(self):
        with self._thread_cond:
            async with self._async_cond:
                self._sema_value = float('inf')
                self._thread_cond.notify_all()
                self._async_cond.notify_all()

    def acquire_threadsafe(self):
        if self._opened:
            while True:
                with self._thread_cond:
                    value = self._sema_value
                    if value > 0:
                        self._sema_value -= 1
                        break
                    self._thread_cond.wait()
        return False

    async def acquire(self):
        if self._opened:
            while True:
                with self._thread_cond:
                    async with self._async_cond:
                        value = self._sema_value
                        if value > 0:
                            self._sema_value -= 1
                            break
                        await self._async_cond.wait()
        return False

    async def prepare(self):
        assert self._stopped
        self._async_cond = asyncio.Condition()
        self._stopped = False

    async def run(self):

        async_sleep = asyncio.sleep
        block_grp = self.parent.block_grp
        config = self.parent.config
        max_speed = config.max_speed
        fraction = 0
        if max_speed is not None:
            self._opened = True
        while True:
            if self._stopped:
                break
            await async_sleep(config.interval)

            # 刷新总的下载块实时传输速率
            block_grp.usage_info.refresh()

            # 当最大下载速度配置有变化后则响应相应的速度限速开关
            if config.max_speed != max_speed:
                # 最大速度限制参数被修改
                max_speed = config.max_speed
                if max_speed is None:
                    self._opened = False
                    await self._release_all()
                else:
                    self._opened = True
                    fraction = 0

            # 如果限制的下载速率就处理信号量
            if max_speed is not None:
                value = config.max_speed * config.interval / 8196

                # 由于下载客户端以单次读数据粒度进行限速，所以为了更细化的限速
                # 对计算出来的信号量粒度小数保留下来留给下次累加。
                fraction += value % 1
                value = int(value)
                if fraction >= 1:
                    value += 1
                    fraction -= 1
                with self._thread_cond:
                    async with self._async_cond:
                        self._sema_value = value
                        self._thread_cond.notify_all()
                        self._async_cond.notify_all()

    async def close(self):
        pass

    async def pause(self):
        if not self._stopped:
            self._stopped = True
            self._opened = False
            await self._release_all()

    def __repr__(self):
        return f'<SpeedAdjuster max_speed={self.parent.config.max_speed} future={self._future}>'

    def info_getter(self):
        return {
            'value': self._sema_value
        }


class FileTempData(Handler):
    """ 下载文件缓冲和保存的IO读写器。

    负责工作：
        1. 文件缓冲和写入
        2. 下载状态的保存
    """

    name = 'file_data'

    def __init__(self):
        self._buffers = defaultdict(list)
        self._counter = 0
        self._unreleased = None
        self._lock = threading.RLock()
        self._stopped = True

    async def saving_state(self):
        """ 保存当前下载状态。

        以cfg的文件形式保存当前下载配置以备文件下载状态的恢复。
        """
        dumpy = self.parent.dumps()
        async with h.aio.open(f'{self.parent.file.pathname}{self.parent.config.downloading_ext}.cfg', mode='w') as f:
            await f.write(json.dumps(dumpy))

    async def _release(self):
        buffers = self._buffers
        counter = self._counter
        self._counter = 0
        self._buffers = defaultdict(list)
        return await self._unreleased.put((counter, buffers))

    def store_threadsafe(self, data):
        """ 线程安全保存临时下载数据。"""
        with self._lock:
            block = _lookup_block()
            self._buffers[block.progress].append(data)
            self._counter += len(data)
            if self.parent.config.buffer_size <= self._counter:
                await_coroutine_threadsafe(self._release())

    async def store(self, data):
        """ 缓冲传输数据。

        当缓冲的数据超过了buffer_size，将对缓冲进行释放写入文件。

        Args:
            data: 要被缓冲的传输数据
        """
        block = _lookup_block()
        self._buffers[block.progress].append(data)
        self._counter += len(data)
        if self.parent.config.buffer_size <= self._counter:
            await self._release()

    async def prepare(self):
        assert self._stopped
        self._unreleased = asyncio.Queue()
        self._stopped = False

    async def run(self):
        unreleased = self._unreleased
        file = self.parent.file
        filepath = f'{file.pathname}{self.parent.config.downloading_ext}'

        # 通过下载块是否有walk_length的情况来判断是否需要重写文件。
        if not self.parent.block_grp.done_length():
            async with h.aio.open(f'{file.pathname}{self.parent.config.downloading_ext}', mode='wb') as fd:
                if file.size is not None:
                    await fd.seek(file.size - 1)
                    await fd.write(b'\x00')

        async with h.aio.open(filepath, mode='rb+') as fd:
            while True:
                result = await unreleased.get()
                if result is None:
                    unreleased.task_done()
                    break
                counter, buffers = result
                for pg, lines in buffers.items():
                    await fd.seek(pg.begin + pg.done_length)
                    await fd.writelines(lines)
                    pg.done(sum([len(line) for line in lines]))

                # 删除引用，尽快回收垃圾
                del lines
                del result
                del buffers
                await self.saving_state()
                unreleased.task_done()

    async def pause(self):
        if not self._stopped:
            self._stopped = True
            await h.client_worker.join()
            await self._release()
            await self._unreleased.put(None)

    async def close(self):
        pass

    def info_getter(self):
        return {
            'size': self._counter,
            'ready': 1
        }

    def __repr__(self):
        return f'<FileTempData counter={self._counter} future={self._future}>'


class AIOReaderWriter(Handler):
    """ AIO读写工作线程。

    为了避免IO的文件读写阻塞影响下载工作线程，该处理器实现异步文件IO读写方法

    负责工作：
        1. 管理IO读写线程
    """
    name = 'aio'

    def __init__(self):
        self._executor = None
        self._writers = set()

    async def prepare(self):
        self._executor = ThreadPoolExecutor(
            max_workers=1, thread_name_prefix=f'BufferWriter {self.parent.file.name}')

    @asynccontextmanager
    async def open(self, file, mode='r', *args, **kwargs):
        """ 异步打开文件。

        Args:
            file: 参见io.open()方法参数file
            mode: 参见io.open()方法参数mode
            args: 参见io.open()方法参数的列表参数
            kwargs: 参见io.open()方法参数字典参数

        Returns:
            异步文件对象AsyncIOFile，对耗时IO文件操作进行异步定义。
        """
        def async_open():
            return open(file, mode, *args, **kwargs)

        executor = self._executor
        assert executor
        loop = asyncio.get_running_loop()
        fd = await loop.run_in_executor(executor, async_open)
        aiofile = AIOFile(executor, fd, loop=loop)
        self._writers.add(aiofile)
        yield aiofile
        # 关闭文件
        await loop.run_in_executor(executor, fd.close)
        self._writers.remove(aiofile)

    async def run(self):
        pass

    async def close(self):
        for handler in h.iter_all():
            if handler != self:
                await handler.join()
        self._executor.shutdown(False)

    async def pause(self):
        pass


class AIOFile:
    """ 异步文件读写对象。

    由AIOReaderWriter的工作线程处理的异步读写对象，将耗时的IO读写由工作线程执行。
    """
    _async_attr = frozenset(
        {'read', 'readline', 'readlines', 'write', 'writeline',
         'writelines', 'seek', 'flush', 'truncate'})

    def __init__(self, executor, fd, loop=None):
        self._executor = executor
        self._fd = fd
        self._loop = loop

    def __getattr__(self, item):
        func = getattr(self._fd, item)
        if item in self._async_attr:
            def ready(*args, loop=None, **kwargs):
                if loop is None:
                    loop = asyncio.get_running_loop()

                if kwargs:
                    handler = partial(getattr(self._fd, item), **kwargs)
                else:
                    handler = getattr(self._fd, item)
                fut = loop.run_in_executor(self._executor, handler, *args)
                return fut
            func = ready

        return func

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def __repr__(self):
        return f'<AIOFile {self._fd}>'



import time
from collections import deque
import asyncio
from contextlib import contextmanager
import threading
from concurrent import futures


class UsageInfo:
    """ 用于记录使用信息。
    其中包括使用时长和使用速率。
    """

    __slots__ = '_fetch_length', '_previous_length', '_previous_time', '_start_time', '_moving_avg', 'rate'

    def __init__(self, fetch_length):
        self._fetch_length = fetch_length

        self._previous_length = fetch_length()
        self._start_time = time.time()
        self._previous_time = self._start_time
        self._moving_avg = deque([0 for _ in range(8)])
        self.rate = 0

    def reset(self):
        self._start_time = time.time()
        self._previous_time = self._start_time
        self._moving_avg = deque([0 for _ in range(8)])
        self.rate = 0

    def timelength(self):
        return time.time() - self._start_time

    def refresh(self):
        cur_time = time.time()
        cur_length = self._fetch_length()
        diff_time = cur_time - self._previous_time
        diff_length = cur_length - self._previous_length

        self._previous_length = cur_length
        self._previous_time = cur_time
        speed = diff_length / (diff_time or float('inf'))

        self._moving_avg.pop()
        self._moving_avg.appendleft(speed)

        self.rate = sum(self._moving_avg) / 8


def update_range_field(range_filed, target_range):
    """ 更新范围域。

    通过{}包含域定义，如{begin}
    可选域定义：
        begin: 范围开头。
        end: 范围结尾，不包括该值本身。
        end_with: 范围结尾，包括该值本身。
        length: 范围长度。

    示例：target_range=(2, 99)
    域值更新为：begin=2, end=99, end_with=98, length=97

    Args:
        range_filed: 范围域定义，可更新域有begin,end,end_with,length
        target_range: 要更新的范围值

    Returns:
        更新域值后的结果。
    """
    target_begin, target_end = target_range

    begin = target_begin
    if target_end is None or target_end == float('inf'):
        end = ''
        end_with = ''
        length = ''
    else:
        end = target_end
        if target_end > 0:
            end_with = target_end - 1
        else:
            end_with = ''
        length = end - begin
    return range_filed.format(
        begin=begin,
        end=end,
        end_with=end_with,
        length=length)


class _ExecutorEventLoopFuture:
    """ 在Executor中安全运行run_forever的事件循环的Future。 """
    def __init__(self, task_fut, loop_fut):
        self._loop = loop_fut
        self._task = task_fut

    def __await__(self):
        yield from asyncio.wrap_future(self._task)

    def __iter__(self):
        yield from asyncio.wrap_future(self._task)

    def get_loop(self):
        return self._loop.result()

    async def aget_loop(self):
        return await asyncio.wrap_future(self._loop)

    def join(self):
        return self._task.result()

    async def ajoin(self):
        return await asyncio.wrap_future(self._task)

    result = join

    aresult = ajoin

    def close(self):
        loop = self.get_loop()
        return loop.call_soon_threadsafe(loop.stop)

    async def aclose(self):
        loop = await self.aget_loop()
        fut = loop.call_soon_threadsafe(loop.stop)
        return await asyncio.wrap_future(fut)

    def add_done_callback(self, __fn):
        return self._task.add_done_callback(__fn)


def forever_loop_in_executor(executor, loop=None):
    """ 在concurrent.futures.thread.ThreadPoolExecutor中运行异步事件循环。
    该事件循环线程只能使用loop.stop()方法停止。"""
    def _run():
        nonlocal loop, future_loop
        if loop is None:
            try:
                loop = asyncio.get_event_loop()
                if loop.is_closed():
                    raise RuntimeError('new loop')
            except RuntimeError:
                loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        future_loop.set_result(loop)
        try:
            loop.run_forever()
        finally:
            try:
                cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                loop.close()

    future_loop = futures.Future()
    task_fut = executor.submit(_run)
    return _ExecutorEventLoopFuture(task_fut, future_loop)


def cancel_all_tasks(loop):
    """ 关闭循环中剩余的所有任务。 """
    # source from asyncio.runners._cancel_all_tasks

    to_cancel = asyncio.tasks.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(
        asyncio.tasks.gather(*to_cancel, loop=loop, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'unhandled exception during asyncio.run() shutdown',
                'exception': task.exception(),
                'task': task,
            })


import os


class File:
    __slots__ = 'name', 'path', 'size'

    def __init__(self, path, name, size):
        """
        :param
            path    : 文件路径（不包括文件名）
            name    : 文件名称
            size    : 文件大小
        """
        self.name = name
        self.path = path
        self.size = size

    @property
    def extension(self):
        return os.path.splitext(self.name)[-1]

    @property
    def pathname(self):
        return os.path.join(self.path, self.name)

    def number_name(self, number):
        just_name, ext = os.path.splitext(self.name)
        return f'{just_name}({number}){ext}'

    def __repr__(self):
        return f'<File {self.pathname}>'

    def dumps(self):
        return {
            'path': self.path,
            'name': self.name,
            'size': self.size,
        }




import json
import mimetypes
import os
import asyncio
from concurrent.futures.thread import ThreadPoolExecutor
from .utils import forever_loop_in_executor

from .download import Downloader
from .client import get_policy
from .uri import URIs
from .progress import Progress
from .error import MaxRetriesExceeded
from .block import BlockGroup
from .file import File
from .request import Request


__all__ = [
    'dlopen',
]


def dlopen(request, handlers=None, *, do_async=True, executors=None):
    """ 打开下载请求Request对象，并构造返回Downloader。
    Args:
        request: 下载请求对象或下载配置文件路径。
        handlers: 添加的Handler对象列表，仅适用于打开下载配置文件
        do_async: 是否使用异步打开
        executors: 使用指定的concurrent.futures.thread打开，默认新创线程执行。
    """
    async def open_request():
        # 打开请求Request对象
        client_policy = request.client_policy
        if client_policy is None:
            client_policy = get_policy()

        uris = URIs.load_from_source_uris(request.uris)
        progress = Progress((0, None))
        source_uri = None
        resp = None
        exceptions = []

        max_retries = request.max_retries
        if request.max_retries is None:
            max_retries = float('inf')

        while True:
            for source_uri in uris:
                try:
                    client_cls = client_policy.get_solution(source_uri.protocol)
                    resp = await client_cls.dlopen(
                        source_uri, progress, **source_uri.kwargs)
                except BaseException as err:
                    exceptions.append(err)
                    max_retries -= 1
                    if max_retries < 0:
                        raise MaxRetriesExceeded(f'max_retries: {request.max_retries}', exceptions)
                else:
                    break
            else:
                continue
            break

        source_uri.set_response(resp)
        path, name = os.path.split(request.file_path)
        if not name:
            name = source_uri.path.rsplit('/', 1)[-1]
            if not name:
                ext = mimetypes.guess_extension(resp.content_type)
                name = f'{source_uri.hostname}{ext or ""}'
        size = resp.length

        block_grp = BlockGroup(request.chunk_size, size)
        block_grp.insert((0, size))
        opts = request.opts
        opts.update(dict(
            client_policy=client_policy,
            resume_capability=resp.resume_capability,
        ))
        return Downloader(
            File(path, name, size),
            uris,
            block_grp,
            **opts
        )

    async def open_cfg():
        # 打开下载配置文件
        file = request
        if not os.path.isfile(file):
            raise FileNotFoundError(f'下载数据配置文件{file}未找到。')
        with open(file, mode='r') as fd:
            dumpy_json = fd.read()
        dumpy = json.loads(dumpy_json)
        return Downloader.loads(dumpy, handlers)

    async def do_open():
        if isinstance(request, Request):
            return await open_request()
        else:
            return await open_cfg()

    def callback(fut):
        executors.shutdown(False)

    new_executor = False
    if executors is None:
        executors = ThreadPoolExecutor(
            max_workers=1, thread_name_prefix='Nbdler.dlopen() Worker')
        new_executor = True

    exec_fut = forever_loop_in_executor(executors)
    if new_executor:
        exec_fut.add_done_callback(callback)

    loop = exec_fut.get_loop()
    if do_async:
        def done_stop_loop(fut):
            nonlocal exec_fut
            exec_fut.close()

        future = asyncio.wrap_future(
            asyncio.run_coroutine_threadsafe(do_open(), loop=loop))
        future.add_done_callback(done_stop_loop)
        result = _AsyncDownloadOpenContextManager(future)
    else:
        result = exec_fut.result()
        exec_fut.close()
    return result


class _AsyncDownloadOpenContextManager:
    __slots__ = '_future', '_result'

    def __init__(self, future):
        self._future = future
        self._result = None

    def __await__(self):
        return self._future.__await__()

    def __iter__(self):
        return self.__await__()

    async def __aenter__(self):
        self._result = await self._future
        return await self._result.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._result.__aexit__(exc_type, exc_val, exc_tb)




from concurrent.futures.thread import ThreadPoolExecutor
from nbdler.handler import (
    SpeedAdjuster,
    AIOReaderWriter,
    BlockSlicer,
    FileTempData,
    ClientWorker,
    URIStatusManager,
    GatherException,
    h, Handlers)
from .client import get_policy, ClientPolicy
from .version import VERSION
from .utils import forever_loop_in_executor
from traceback import format_exc
import weakref
import warnings
import asyncio
import os

__all__ = (
    'Downloader',
)


class DownloadConfigure:
    ADJUSTABLE = frozenset(
        {'max_concurrent', 'max_speed', 'buffer_size', 'timeout', 'interval', 'client_policy'})

    def __init__(self, resume_capability, max_concurrent, chunk_size, buffer_size, timeout=10,
                 max_speed=None, downloading_ext='.downloading', interval=0.5, client_policy=None, **kwargs):

        self.version = VERSION
        self.resume_capability = resume_capability
        self.max_concurrent = max_concurrent
        self.chunk_size = chunk_size
        self.buffer_size = buffer_size
        self.timeout = timeout
        self.interval = interval
        self.max_speed = max_speed
        self.downloading_ext = downloading_ext
        self.client_policy = client_policy
        self.kwargs = kwargs

    def set(self, **kwargs):
        """ 设置配置。
        Args:
            **kwargs:
                max_concurrent: 最大并发数
                max_speed: 最大速度限制
                buffer_size: 最大文件缓冲大小
                timeout: 客户端连接接收超时时间
                interval: 速度调节间隙
                client_policy: 客户端处理策略
        """
        attrs = set(kwargs).intersection(DownloadConfigure.ADJUSTABLE)
        for attr in attrs:
            self.__setattr__(attr, kwargs[attr])

    def dumps(self):
        opts = dict(self.__dict__)
        client_policy = self.client_policy
        opts['client_policy'] = dict(client_policy)
        opts.update(opts.pop('kwargs'))
        return opts

    @classmethod
    def loads(cls, dumpy):
        config = cls(**dumpy)
        if not isinstance(config.client_policy, ClientPolicy):
            config.client_policy = get_policy(**config.client_policy)
        return config

    def __repr__(self):
        return (f'<DownloadConfigure version={self.version} max_concurrent={self.max_concurrent} '
                f'resume_capability={self.resume_capability}>')


class Downloader:
    def __init__(self, file, uris, block_grp, *, handlers=None, **kwargs):

        self.file = file
        self.uris = uris
        self.block_grp = block_grp
        self.config = DownloadConfigure.loads(kwargs)

        self._executor = None

        self._loop = None
        self._future = None
        self._closed = False
        self._handlers = Handlers()

        if handlers is None:
            handlers = []

        buildin_handlers = [
            ClientWorker,
            SpeedAdjuster,
            FileTempData,
            AIOReaderWriter,
            BlockSlicer,
            GatherException,
            URIStatusManager,
        ]
        handlers.extend(buildin_handlers)
        for handler in handlers:
            if handler.name in self._handlers:
                continue
            if isinstance(handler, type):
                handler = handler()

            handler.add_parent(weakref.proxy(self))
            self._handlers[handler.name] = handler

    def exceptions(self, exception_type=None, *, just_new_exception=True):
        """ 线程安全获取异常

        以生成器的形式获取内部发生的异常，当下载任务暂停或者完成后将中断生成器的迭代。
        使用方式：
            for exception in dl.exceptions():
                do_some_works(exception)

        Args:
            exception_type: 指定异常类型，可选ClientError、HandlerError。默认None则获取所有异常。
            just_new_exception: 是否忽略当前时间前的旧异常，仅返回之后的新异常。

        Yields:
            内部出现的client或handler异常对象。
        """
        yield from self._handlers.exception.acquire_threadsafe(
            exception_type, just_new_exception=just_new_exception)

    def aexceptions(self, exception_type=None, *, just_new_exception=True):
        """ 异步返回异常错误。 具体参见exceptions()方法。
        使用方式：
            async for exception in dl.aexceptions():
                do_some_works(exception)
        """
        return self._handlers.exception.acquire(
            exception_type, just_new_exception=just_new_exception)

    async def astart(self):
        """ 在当前事件循环中运行下载器。"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')
        loop = asyncio.get_running_loop()
        self._loop = loop
        if self.block_grp.is_done_finished():
            raise RuntimeError('download is completed.')

        self._future = loop.create_future()

        async def handler_worker(hd):
            try:
                return await hd.start()
            except BaseException as err:
                h.exception.handler_error(err)
                self.pause(0)

        with h.enter(self._handlers, loop):
            self.block_grp.activate()
            # prepare()
            await self._handlers.prepare()
            # start()
            result = await asyncio.gather(
                *[handler_worker(handler) for handler in h.iter_all()]
            )
            # join()
            await self._handlers.join()
            self.block_grp.deactivate()

        self._future.set_result(result)

    def start(self, *, loop=None):
        """ 在指定的循环中运行下载器。

        若loop=None不指定事件循环，那么将创建新的线程作为下载器的事件循环。

        Args:
            loop: 指定事件循环运行下载器

        Returns:
            返回下载器运行的concurrent.future.Future对象
        """

        if self._closed:
            raise RuntimeError('Downloader is already closed.')

        if self.block_grp.is_finished():
            raise RuntimeError('download is already finished.')

        if self._loop is not None:
            loop = self._loop

        if loop is None:
            def cb(f):
                nonlocal executor
                executor.shutdown(False)

            executor = ThreadPoolExecutor(
                max_workers=1, thread_name_prefix=f'Downloader {self.file.name} {self.file.size}')
            exec_fut = forever_loop_in_executor(executor)
            exec_fut.add_done_callback(cb)
            self._executor = executor
            loop = exec_fut.get_loop()

        fut = asyncio.run_coroutine_threadsafe(self.astart(), loop=loop)
        self._loop = loop
        return fut

    async def apause(self):
        """ 异步暂停等待。"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')
        result = await self._await_loopsafe(self._handlers.pause())
        await self.ajoin()
        return result

    async def aclose(self):
        """ 异步关闭下载器。"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')

        if not self._future.done():
            raise RuntimeError('cannot close a running Downloader.')
        result = await self._await_loopsafe(self._handlers.close())
        await self.ajoin()
        self._closed = True

        if self._executor:
            self._loop.call_soon_threadsafe(self._loop.stop)

        # 若文件已完毕，去除.downloading后缀
        if self.block_grp.is_done_finished():
            file = self.file
            filepath = f'{file.pathname}{self.config.downloading_ext}'
            start_filepath = file.pathname
            target_filepath = start_filepath
            postfix = 0
            while True:
                try:
                    os.rename(filepath, target_filepath)
                except FileExistsError:
                    postfix += 1
                    target_filepath = os.path.join(file.path, file.number_name(postfix))
                else:
                    if postfix != 0:
                        file.name = file.number_name(postfix)
                    break

            # 删除下载配置文件
            os.unlink(f'{start_filepath}{self.config.downloading_ext}.cfg')
        return result

    async def ajoin(self):
        """ 异步等待下载器结束。"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')
        return await self._await_loopsafe(self._future)

    async def _await_loopsafe(self, *coros_or_futures):
        """ 事件循环安全的异步等待。

        Args:
            *coros_or_futures: coroutine或future对象列表。

        Returns:
            返回coros_or_futures的返回结果列表。
        """
        current_loop = asyncio.get_running_loop()
        loop = self._loop
        if loop is None:
            loop = current_loop

        async def _execute_loop():
            with h.enter(self._handlers):
                r = await asyncio.gather(*coros_or_futures)
                return r
        fut = asyncio.run_coroutine_threadsafe(_execute_loop(), loop)
        result = await asyncio.wrap_future(fut)

        return result

    def _call_threadsafe(self, coroutine, timeout=None):
        """ 下载器的异步操作线程安全化。
        Args:
            coroutine: 异步操作协程
            timeout: 超时等待事件

        Returns:
            当timeout=0时，返回concurrent.future.Future对象，
            否则，协程coroutine的执行结果或抛出超时异常。
        """
        loop = self._loop
        assert loop
        future = asyncio.run_coroutine_threadsafe(coroutine, loop)
        if timeout == 0:
            return future
        return future.result(timeout)

    def pause(self, timeout=None):
        """ 线程安全暂停下载器。具体参见apause方法"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')
        return self._call_threadsafe(self.apause(), timeout=timeout)

    def close(self, timeout=None):
        """ 线程安全关闭下载器。具体参见aclose方法"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')
        return self._call_threadsafe(self.aclose(), timeout=timeout)

    def join(self, timeout=None):
        """ 线程安全等待下载器。具体参见ajoin方法"""
        if self._closed:
            raise RuntimeError('Downloader is already closed.')
        return self._call_threadsafe(self.ajoin(), timeout=timeout)

    def dumps(self):
        dumpy = {
            'config': self.config.dumps(),
            'file': self.file.dumps(),
            'uris': self.uris.dumps(),
            'block_grp': self.block_grp.dumps(),
        }
        return dumpy

    @classmethod
    def loads(cls, dumpy, handlers=None):
        from nbdler.uri import URIs
        from nbdler.file import File
        from nbdler.block import BlockGroup

        uris = URIs.loads(dumpy['uris'])
        file = File(**dumpy['file'])
        block_grp = BlockGroup.loads(dumpy['block_grp'])
        return cls(file, uris, block_grp, handlers=handlers, **dumpy['config'])

    transfer_rate = property(lambda self: self.block_grp.transfer_rate)

    average_speed = property(lambda self: self.block_grp.average_speed)

    walk_length = property(lambda self: self.block_grp.walk_length)

    done_length = property(lambda self: self.block_grp.done_length)

    remaining_length = property(lambda self: self.block_grp.remaining_length)

    remaining_time = property(lambda self: self.block_grp.remaining_time)

    percent_complete = property(lambda self: self.block_grp.percent_complete)

    is_walk_finished = property(lambda self: self.block_grp.is_walk_finished)

    is_done_finished = property(lambda self: self.block_grp.is_done_finished)

    def is_finished(self):
        """ 返回文件是否下载完毕。"""
        return self.block_grp.is_finished() and (not self._future or self._future.done())

    def set_config(self, **kwargs):
        """ 配置下载器。参见DownloadConfigure.set()方法。"""
        self.config.set(**kwargs)

    def __repr__(self):
        running = False
        if self._future is not None and not self._future.done():
            running = True
        return f'<Downloader filename={self.file.name} running={running} closed={self._closed}>'

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.aclose()

    def __del__(self, _warnings=warnings):
        if not self._closed:
            self.close()





# -*- coding: UTF-8 -*-

from nbdler.uri import SourceURI
from typing import Sequence, Mapping, AnyStr, Union, Optional
from nbdler.client import ClientPolicy
import bisect


class RequestConfig:
    def __init__(self, *, file_path: AnyStr,
                 max_concurrent: int=5,
                 buffer_size: int=20*1024*1024,
                 chunk_size: int=64*1024,
                 max_retries: Optional[int]=3,
                 timeout: Optional[int]=10,
                 client_policy: Optional[ClientPolicy]=None,
                 **kwargs):
        """
        Args:
            file_path: 目标文件的路径名称
            max_concurrent: 最大并发数
            buffer_size: 最大请求次数仅用于dlopen的时候尝试的次数。
            chunk_size: 下载客户端的最大线程数量。
            max_retries: 内存缓冲的最大值。
            timeout: 下载客户端的超时参数。None一般代表着时无限阻塞链接。
            client_policy: 指定客户端处理策略，默认策略由 nbdler.client.__init__._DEFAULT_POLICY 指定
            **kwargs: 保存提供的额外参数，以在后续提供给下载客户端来让客户端进行选择调整。
                      - downloading_ext: 下载中文件扩展名
                      - interval: 心跳刷新间隔
        """
        self.file_path = file_path
        self.max_concurrent = max_concurrent
        self.chunk_size = chunk_size
        self.max_retries = max_retries
        self.buffer_size = buffer_size
        self.timeout = timeout
        self.client_policy = client_policy
        self.handlers = []
        self.kwargs = kwargs

    def add_handler(self, *handlers):
        """ 添加或更新下载处理器。
        Args:
            *handlers: 继承nbdler.handler.Handler类的处理器列表参数，
                       通过名称标识处理器，类变量name作为处理器唯一标识。
                       与内置处理器重名则替换内置处理器。
                       内置处理器有:
                       - client_worker: 客户端调配控制器
                       - slicer: 下载块切片器
                       - speed_adjuster: 速度调节器，用于限速和实时速率刷新
                       - uri_mgr: 下载源状态管理器
                       - exception: 异常收集处理器
                       - file_data: 文件缓冲区
                       - aio: 异步文件读写工作线程
        """
        for handler in handlers:
            bisect.insort(self.handlers, handler)

    def __repr__(self):
        return f'<BaseRequest {self.file_path}>'


class Request(RequestConfig):

    def __init__(self, uri=None, headers=None, cookies=None, proxies=None,
                 max_conn=None, range_field=None, name='main', *, file_path, **kwargs):
        """
        Args:
            uri,headers,cookies,proxies,max_conn,range_field,name: 参考put()方法。
            **kwargs: 参考RequestConfig的__init__()方法
        """
        super(Request, self).__init__(file_path=file_path, **kwargs)
        self._uris = []
        if uri:
            kwargs = {k: kwargs[k] for k in set(kwargs).difference(self.__dict__)}
            self.put(uri, headers, cookies, proxies=proxies,
                     max_conn=max_conn, range_field=range_field, name=name, **kwargs)

    def put(self, uri: AnyStr,
            headers: Optional[Union[Sequence, Mapping[str, str]]]=None,
            cookies: Optional[Mapping[str, str]]=None,
            proxies: Optional[Mapping]=None,
            max_conn: Optional[int]=None,
            range_field: Optional[Mapping[str, str]]=None,
            name: Optional[str]=None,
            **kwargs):
        """ 添加下载源。
        Args:
            uri: URI链接
            headers: 请求头
            cookies: 下载源请求传递的Cookie，要求传递字典dict类型
            proxies: 代理服务器，带auth例子：{'http': 'http://user:pass@some.proxy.com'}
            max_conn: 最大连接数
            range_field: 范围请求定义，要求提供字典类型，如 {'Range': 'bytes={begin}-{end_with}'}
            name: 下载源名称，仅用于标记，默认不提供，系统自动编号
            **kwargs: 允许根据下载源参数指定客户端的特定操作。
                - trust_env:          使用系统代理

        Returns:
            返回未经编号的下载源。
        """
        uri = uri.strip()
        src_url = SourceURI(None, uri, headers, cookies, proxies, max_conn, range_field, name, **kwargs)
        self._uris.append(src_url)
        return src_url

    @property
    def opts(self):
        """ 返回请求中的配置字典信息。

        Returns:
            返回配置信息字典，具体键值参考RequestConfig。
        """
        opts = {k: v for k, v in self.__dict__.items() if not k.startswith('_')}
        opts.update(opts.pop('kwargs'))
        return opts

    @property
    def uris(self):
        return self._uris

    def __repr__(self):
        return f'<Request {self._uris and self._uris[0].uri}>'

    def dumps(self):
        return {
            'config': self.opts,
            'uris': [uri.dumps() for uri in self._uris]
        }

    @classmethod
    def loads(cls, dumpy):
        request = cls(**dumpy['config'])
        for uri in dumpy['uris']:
            request.put(**uri)





import aiohttp
import asyncio
from urllib.parse import urlunparse, urlparse
from nbdler.uri import URIResponse
from .base_http import BaseHTTPClient, content_range_fullsize, content_type_mimetype
from traceback import format_exc
from nbdler.handler import h
import logging
import nbdler

log = logging.getLogger(__name__)


class AIOHTTPClient(BaseHTTPClient):
    TIMEOUT = 10

    async def connect(self):
        session = self.session
        source_uri = self.source_uri
        proxies = source_uri.proxies or {}
        proxy = None
        if not proxies.get(source_uri.scheme):
            if source_uri.kwargs.get('trust_env', False):
                for scheme, proxy_info in aiohttp.helpers.proxies_from_env().items():
                    if scheme == source_uri.scheme:
                        proxy = str(proxy_info.proxy)
                        proxy_auth = proxy_info.proxy_auth
                        if proxy_auth is not None:
                            # 将代理验证添加入代理链接
                            username = proxy_auth.login
                            password = proxy_auth.password
                            proxy_parse = urlparse(str(proxies))
                            scheme, netloc, path, params, query, fragment = list(proxy_parse)
                            netloc = f'{username}:{password}@{netloc}'
                            proxy = urlunparse([scheme, netloc, path, params, query, fragment])
                        break

        cookies = source_uri.cookies
        uri, headers = self._build_uri_headers()

        timeout = self.kwargs.get('timeout', None) or AIOHTTPClient.TIMEOUT
        timeout = aiohttp.ClientTimeout(sock_connect=timeout, sock_read=timeout)

        try:
            resp = await session.get(
                uri,
                headers=headers,
                cookies=cookies,
                proxy=proxy,
                timeout=timeout,
            )
        except (aiohttp.ClientError, asyncio.TimeoutError) as error:
            raise nbdler.error.TimeoutError(f"{uri}") from error
        except BaseException as error:
            log.debug(f'{error}', format_exc())
            raise nbdler.error.FatalError() from error
        else:
            total_length = content_range_fullsize(resp.headers.get('content-range'))
            response = URIResponse(str(resp.url), list(resp.headers.items()), resp.status, resp.reason,
                                   total_length, content_type_mimetype(resp.headers.get('content-type')),
                                   self.progress.range, resp.status == 206)

        if self.resume_capability is None:
            if resp.status not in (206, 200):
                raise nbdler.error.FatalError(f"[{resp.status} {resp.reason}] '{resp.url}'")
            self.resume_capability = resp.status == 206

        elif self.resume_capability is True:
            if not resp.status == 206:
                raise nbdler.error.FatalError(f"[{resp.status} {resp.reason}] '{resp.url}'")

        self.session = session
        self.resp = resp
        return response

    async def fetch(self):
        session, resp = self.session, self.resp
        pg = self.progress

        speed_adjuster = h.speed_adjuster
        slicer = h.slicer
        uri_mgr = h.uri_mgr
        file_data = h.file_data

        pg.start()

        uri_mgr.success(resp)

        receive_data = resp.content.read
        data = b''
        while True:
            if self._closed:
                break

            await speed_adjuster.acquire()
            await slicer.response()

            previous_len = len(data)
            remain_len = pg.total_length - pg.walk_length
            try:
                if remain_len >= 8192:
                    data += await receive_data(8192)
                elif remain_len > 0:
                    data += await receive_data(remain_len)
                else:
                    break
            except asyncio.TimeoutError as err:
                uri_mgr.timeout(err)
                break
            except BaseException as err:
                uri_mgr.fatal(err)
                break

            walk_len = len(data) - previous_len
            if not walk_len:
                if resp.content_length is None:
                    pg.set_walk_finish()
                break

            pg.walk(walk_len)

            if pg.walk_length >= pg.total_length:
                break
            elif len(data) >= 65536:  # 64 KB
                await file_data.store(data)
                data = b''
        if data:
            await file_data.store(data)

        pg.stop()

    async def run(self):
        if self.resp:
            self.close()

        await h.slicer.response()
        try:
            resp = await self.connect()
        except nbdler.error.UriError as err:
            h.uri_mgr.fatal(err)
            raise
        else:
            h.uri_mgr.success(resp)

            # self.validate_token(resp)
            if not self._closed:
                await self.fetch()

    def close(self):
        session = self.session
        resp = self.resp
        self.session = None
        self.resp = None
        if resp:
            resp.release()
            resp.close()

    @classmethod
    async def dlopen(cls, source, progress, **kwargs):
        async with ClientSession() as session:
            async with cls(session, source, progress, None, **kwargs) as cli:
                resp = await cli.connect()
                size = resp.length
                progress._range = (0, size)

        return resp


NAME = 'aiohttp'
PROTOCOL_SUPPORT = ('http', 'https')
ASYNC_EXECUTE = True

ClientHandler = AIOHTTPClient


class ClientSession(aiohttp.ClientSession):
    async def close(self) -> None:
        await super().close()

        # doc: https://docs.aiohttp.org/en/latest/client_advanced.html#graceful-shutdown
        # 会话关闭强制等待避免异常
        await asyncio.sleep(0.25)





import asyncio


class AbstractClient:
    """ 抽象客户端

    Class Variable:
        NAME: 客户端名称，作为客户端的唯一标识
        PROTOCOL_SUPPORT: 客户端支持处理的协议, protocol/scheme
        ASYNC_EXECUTE: 指定客户端的是否为异步实现
        TIMEOUT: 客户端默认的连接connect,读取read超时参数
    """

    TIMEOUT = 10

    def __init__(self, session,
                 source_uri,
                 progress,
                 resume_capability,
                 **kwargs):
        """
        Args:
            session: 客户端会话
            source_uri: 下载源SourceUri对象
            progress: 请求进度对象Progress
            resume_capability: 是否支持断点续传，若为None则代表不确定，连接后将根据实际情况赋值
        """
        self.source_uri = source_uri
        self.progress = progress
        self.resume_capability = resume_capability
        self.kwargs = kwargs

        self._closed = False
        self.session = session
        self.resp = None

    async def connect(self):
        """ (可定义非异步方法)客户端连接

        Returns:
            UriResponse对象，该对象指定了资源的基本信息。
        """
        raise NotImplementedError

    async def fetch(self):
        """ (可定义非异步方法)客户端循环获取数据 """
        raise NotImplementedError

    async def pause(self):
        """ 客户端暂停 """
        self._closed = True
        raise NotImplementedError

    async def close(self):
        raise NotImplementedError

    async def run(self):
        raise NotImplementedError

    async def __aenter__(self):
        """ 异步with enter.

        进入客户端，准备开始客户端。
        该方法不应该执行非异步的长耗时任务。

        Returns:
            返回自身对象self
        """
        self._closed = False
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """ 异步with exit.

        退出客户端，做必要的链接关闭操作，设置实例变量_closed=True。
        """
        coro_or_result = self.close()
        # 兼容异步关闭链接方法
        if asyncio.iscoroutine(coro_or_result):
            await coro_or_result
        self.session = None
        self._closed = True

    def __enter__(self):
        """ 同步with enter.

        进入客户端，准备开始客户端。
        该方法不应该执行非异步的长耗时任务。

        Returns:
            返回自身对象self
        """
        self._closed = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self.session = None
        self._closed = True

    @classmethod
    def dlopen(cls, source, progress, **kwargs):
        raise NotImplementedError

    def __repr__(self):
        status = 'running'
        if self._closed:
            status = 'closed'
        name = self.run.__globals__['NAME']
        support = self.run.__globals__['PROTOCOL_SUPPORT']
        is_async = self.run.__globals__['ASYNC_EXECUTE']
        return f'<Client {status} name="{name}" ' \
               f'support={support} ' \
               f'async={is_async}>'


def noop():
    """ ignore function. """
    return None


NAME = 'abstract'
PROTOCOL_SUPPORT = ('http', 'https')
ASYNC_EXECUTE = True

ClientSession = noop
ClientHandler = AbstractClient



# -*- coding: UTF-8 -*-
""" 下载块

下载块Chunk结构:

+---------------------------------+
| Chunk                           |
|         +--------+----------+   |
|         | client | progress |   |
|         +--------+----------+   |
+---------------------------------+
| Block [0:25]                    |
| >>>>>-------------------- 05/25 |
+---------------------------------+
|+|+|+|+|+|+|+|+|+|-|-|-|-|-|-|-|-|
|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|  <--------- block
+---------------------------------+

其中: '>' 表示已下载块; '-' 表示未下载块

下载块管理器:

+-------------------------------------------------------------+
| ChunkManager                                                |
|      +--------C1-------+ +---C3----+ +---------C2--------+  |
|      |-|-|-|-|-|-|-|-|-| |-|-|-|-|-| |-|-|-|-|-|-|-|-|-|-|  |
|      |-|-|-|-|-|-|-|-|-| |-|-|-|-|-| |-|-|-|-|-|-|-|-|-|-|  |
|      |-|-|-|-|-|-|-|-|-| |-|-|-|-|-| |-|-|-|-|-|-|-|-|-|-|  |
|      +-----------------+ +---------+ +-------------------+  |
+-------------------------------------------------------------+

文件下载块切片原理:

表述：
    搜索从剩余下载量最多的下载块中切片生成新的下载块。

1. 下载块C1 (总下载块=C1)
+--------------------------------------------------------------------------------------------------------------------+
|                                                C1 [0:100]                                                          |
|        >>>>>>>---------------------------------------------------------------------------------------------        |
+--------------------------------------------------------------------------------------------------------------------+

2. 从下载块B1中切片分出下载块C2 (总下载块=C1+C2)
+--------------------------------------------------------------+-----------------------------------------------------+
|                         C1 [0:54]                            |                   C2 [54:100]                       |
|    >>>>>>>>>>>>>>----------------------------------------    |    >>>-------------------------------------------   |
+--------------------------------------------------------------+-----------------------------------------------------+

3. 从下载块C2中切片分出下载块B3 (总下载块=C1+C2+C3)
+----------------------------------------------------------+-----------------------------+---------------------------+
|                         C1 [0:54]                        |         C2 [54:79]          |        C3 [79:100]        |
|  >>>>>>>>>>>>>>>>>>>>>---------------------------------  |  >>>>>>>>-----------------  |   >>>>-----------------   |
+----------------------------------------------------------+-----------------------------+---------------------------+

4. 从下载块C1中切片分出下载块C4 (总下载块=C1+C2+C3+C4)
+--------------------------------------+----------------------+----------------------------+-------------------------+
|               C1 [0:36]              |      C4 [36:54]      |         C2 [54:79]         |      C3 [79:100]        |
| >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>----- |  >-----------------  |  >>>>>>>>>>>>>>>>>-------- |  >>>>>>>>>>>>>>>>-----  |
+--------------------------------------+----------------------+-------------------------—--+-------------------------+

5. 从下载块C4中切片分出下载块B5 (总下载块=C1+C2+C3+C4+C5)
+--------------------------------------+------------+------------+---------------------------+-----------------------+
|               C1 [0:36]              | C4 [36:46] | C5 [46:54] |         C2 [54:79]        |      C3 [79:100]      |
| >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>-- | >>>>>----- |  >>------  | >>>>>>>>>>>>>>>>>>>>>>--- | >>>>>>>>>>>>>>>>>---- |
+--------------------------------------+------------+------------+---------------------------+-----------------------+

...


"""
from .utils import UsageInfo
from math import ceil
from threading import RLock
from time import time
from .progress import Progress
import bisect


class Chunk:
    __slots__ = 'uri_id', 'begin', 'end'

    def __init__(self, uri_id, begin, end):
        """
        :param
            uri_id     : 源URI的ID
            begin   : 单元的开始索引
            end     : 单元的结束索引
        """
        self.uri_id = uri_id
        self.begin = begin
        self.end = end

    @property
    def length(self):
        return self.end - self.begin

    def __repr__(self):
        return f'<Cell ({self.uri_id})[{self.begin}-{self.end}]>'

    def __iter__(self):
        return iter([self.uri_id, self.begin, self.end])


class Block:
    """ 下载块是对某一刻下载进度的快照，是一个状态量。不具备实时性。
    既然不作为实时数据，那么不需要对其进行强制与实时数据对应。
    那么对于未确定大小的进度对象Progress，即大小为inf时，不必关注其大小，关注的是其下载的量的块记录。
    """
    __slots__ = 'client', 'progress', '_chunk_size', '_range', '_chunks', '_lock'

    def __init__(self, progress, chunk_size, init_chunks=None):
        self.client = None
        self.progress = progress
        self._chunk_size = chunk_size

        begin = progress.begin // chunk_size

        end = 1
        if progress.end not in (float('inf'), None):
            end = int(ceil(progress.end / chunk_size))

        self._range = [begin, end]
        self._chunks = []

        if init_chunks is not None:
            self._chunks = [Chunk(*chunk) for chunk in init_chunks]

        self._lock = RLock()
        self.refresh()

    @property
    def begin(self):
        return self._range[0]

    @property
    def end(self):
        return self._range[1]

    @property
    def length(self):
        try:
            return self._range[1] - self._range[0]
        except TypeError:
            return float('inf')

    @property
    def chunks(self):
        self.refresh()
        return self._chunks

    def current_uri(self):
        return self.client and self.client.source_uri

    def slice(self, request_range):
        resp_range = self.progress.slice(request_range)
        self.refresh()
        return resp_range

    def unused_length(self):
        """ 返回下载块中未处理的块chunk长度。"""
        if self.progress.end in (None, float('inf')):
            return float('inf') if not self.progress.is_walk_finished() else 0
        self.refresh()

        return (not self._chunks and self.length) or self.length - self._chunks[-1].end

    def __getitem__(self, index):
        assert type(index) is int
        for c in self.chunks:
            if c.begin <= index < c.end:
                return c

        assert False

    def refresh(self):
        """ 刷新下载块进度信息。

        由于下载块只是作为快照的形式监控，并不直接由下载客户端接管，所以得到的信息并不是实时的，
        这就需要在获取信息的时候进行刷新信息。
        对外的接口都已进行了刷新数据来保证数据的实时性。
        """
        with self._lock:
            progress = self.progress

            block_begin = self.begin

            cur_uri = self.current_uri()

            cur_walk = progress.walk_length / self._chunk_size
            if not progress.walk_left:
                cur_walk = int(ceil(cur_walk))
            else:
                cur_walk = int(cur_walk)

            cur_done = progress.done_length / self._chunk_size
            if not progress.done_left:
                cur_done = int(ceil(cur_done))
            else:
                cur_done = int(cur_done)

            # 更新块范围
            block_end = progress.end
            if block_end is None:
                block_end = cur_walk + block_begin
            block_end = int(ceil(block_end / self._chunk_size))

            self._range[1] = block_end

            last_chunk = (self._chunks and self._chunks[-1]) or None
            if last_chunk is None:
                if cur_uri is not None:
                    self._chunks.append(Chunk(cur_uri.id, 0, cur_walk))
                else:
                    pass
            else:
                if cur_uri is not None:
                    if last_chunk.uri_id != cur_uri.id:
                        self._chunks.append(Chunk(cur_uri.id, cur_walk, cur_walk))
                    else:
                        last_chunk.end = cur_walk

    def half_unused(self):
        unused_len = self.unused_length()
        put_begin = self.begin + (self.length - unused_len) + int(ceil(unused_len / 2))
        put_end = self.end
        if put_begin == put_end:
            return None
        return put_begin * self._chunk_size, put_end * self._chunk_size

    def request(self, client):
        self.client = client
        return self

    async def __aenter__(self):
        from nbdler.handler import block_context

        assert self.client
        block_context.set(self)
        return await self.client.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        from nbdler.handler import block_context

        self.refresh()
        client = self.client
        await client.__aexit__(exc_type, exc_val, exc_tb)
        self.client = None
        block_context.set(None)

    def dumps(self):
        return {
            'progress': list(self.progress),
            'range': [self.begin, self.end],
            'chunks': [list(c) for c in self._chunks]
        }

    def __repr__(self):
        return f'<Chunk [{self.begin}-{self.end}] {self.progress.percent_complete / 100:.2%} client={self.client}>'

    def __iter__(self):
        return iter([list(self.progress), self._chunk_size, [list(block) for block in self._chunks]])

    def __lt__(self, other):
        return self.begin < other.begin


class BlockGroup:
    """ 下载块管理器。 """
    def __init__(self, chunk_size, total_size, duration=0):
        self._blocks = []
        self.chunk_size = chunk_size
        try:
            self.total_chunk = int(ceil(total_size / chunk_size))
        except (TypeError, OverflowError):
            self.total_chunk = 1

        if total_size is None:
            total_size = float('inf')
        self.total_size = total_size

        self.usage_info = UsageInfo(self.walk_length)

        self._start_time = None
        self._duration = duration

    def transfer_rate(self):
        """ 实时数据传输速率。 """
        return self.usage_info.rate

    def average_speed(self):
        """ 平均数据传输速率。 """
        total_time = self._duration + time() - (self._start_time or time())
        return self.walk_length() / (total_time or float('inf'))

    def walk_length(self):
        """ 已下载字节数。 """
        return sum((v.progress.walk_length for v in self._blocks))

    def done_length(self):
        """ 已缓冲的字节数。"""
        return sum((v.progress.done_length for v in self._blocks))

    def remaining_length(self):
        """ 还剩余字节数。 """
        return self.total_size - self.walk_length()

    def remaining_time(self):
        """ 估计剩余时间。 """
        realtime_rate = self.transfer_rate()
        if not realtime_rate:
            return float('inf')
        return self.remaining_length() / realtime_rate

    def percent_complete(self):
        return self.walk_length() * 100 / self.total_size

    def is_walk_finished(self):
        for b in self._blocks:
            if not b.progress.is_walk_finished():
                return False
        return not self.integrity_check()

    def is_done_finished(self):
        for b in self._blocks:
            if not b.progress.is_done_finished():
                return False
        return not self.integrity_check()

    is_finished = is_done_finished

    def insert(self, put_range):
        """ 插入下载块。
        Args:
            put_range: 插入的快进度范围range
        """
        block = Block(Progress(put_range), self.chunk_size)
        bisect.insort(self._blocks, block)

        return block

    def unfinished_blocks(self):
        return [b for b in self._blocks if not b.progress.is_walk_finished()]

    def activate(self):
        """ 激活下载块映射图。 """
        self._start_time = time()

    def deactivate(self):
        """ 关闭下载块映射图。 """
        self._duration += time() - (self._start_time or time())
        self._start_time = None
        self.usage_info.reset()
        if self.is_walk_finished():
            if self.total_size in (None, float('inf')):
                self.total_size = self.walk_length()

    def integrity_check(self):
        """ 下载块映射图完整性检测。
        如果下载块缺失返回缺失的块，否则返回[]。
        """
        if not self._blocks:
            return [(0, self.total_chunk)]
        missing = []
        prv_end = self._blocks[0].end
        prv_b = None
        for v in self._blocks[1:]:
            v.refresh()
            if v.begin - prv_end > 0:
                # 如果下一个下载块的起点索引比上一个下载块的结束索引要大，说明了这其中缺少了一块。
                missing.append((prv_end, v.begin))
            elif v.begin - prv_end < 0:
                # 如果下一个下载块的起点索引比上一个下载块的结束索引要小，说明这出现了下载块范围交叉。
                raise ValueError(f'完整性校验不通过。冲突：{prv_b} <-> {v}')

            prv_end = v.end
            prv_b = v

        return missing

    def dumps(self):
        return {
            'chunk_size': self.chunk_size,
            'total_size': self.total_size,
            'duration': self._duration,
            'blocks': [b.dumps() for b in self._blocks],
        }

    @classmethod
    def loads(cls, dumpy):
        chunk_size = dumpy['chunk_size']
        block_grp = cls(chunk_size, dumpy['total_size'], dumpy['duration'])
        for block in dumpy['blocks']:
            progress = Progress(*block['progress'])
            block = Block(progress, chunk_size, block['chunks'])
            bisect.insort(block_grp._blocks, block)
        return block_grp

    def __iter__(self):
        """ 迭代返回下载块对象。"""
        return iter([self.chunk_size, self.total_size, [list(block) for block in self._blocks]])

    def __repr__(self):
        return f'<BlockGroup transfer_rate={round(self.transfer_rate() / 1024)} kb/s ' \
               f'percent={round(self.percent_complete(), 2)}%>'



# -*- coding: UTF-8 -*-

# from ..utils.misc import Component
# # from .misc import Timer


class Progress:
    __slots__ = '_range', 'walk_length', 'done_length', '_timer'

    def __init__(self, range, walk_length=0, done_length=0, increment_time=0):
        begin, end = range
        self._range = range

        if done_length != walk_length:
            # 存在下载缓冲未写入文件，下载配置文件可能被非正常关闭。尝试回退下载进度。
            done_length = walk_length

        self.walk_length = walk_length
        self.done_length = done_length

    def is_walk_finished(self):
        return self.walk_length >= self.total_length

    def is_done_finished(self):
        return self.done_length >= self.total_length

    def is_finished(self):
        return self.is_walk_finished() and self.is_done_finished()

    @property
    def range(self):
        return self._range

    @property
    def begin(self):
        return self._range[0]

    @property
    def end(self):
        return self._range[1]

    @property
    def total_length(self):
        try:
            return self._range[1] - self._range[0]
        except TypeError:
            return float('inf')

    @property
    def walk_left(self):
        return self.total_length - self.walk_length

    @property
    def done_left(self):
        return self.total_length - self.done_length

    @property
    def differ(self):
        return self.walk_length - self.done_length

    @property
    def time_length(self):
        return 0

    @property
    def average_speed(self):
        return

    @property
    def percent_complete(self):
        return self.walk_length * 100 / self.total_length

    def walk(self, byte_len):
        self.walk_length += byte_len

    def done(self, byte_len):
        self.done_length += byte_len

    def start(self):
        pass

    def stop(self):
        pass

    def set_walk_finish(self):
        """ 由于存在未指定结尾的情况，也就是未指定下载大小的情况。
        那么当下载完全的情况下，允许强制其以当前下载量作为结尾。
        """
        assert self.end is None
        self._range = (self.begin, self.begin + self.walk_length)

    def reset(self):
        self.walk_length = 0
        self.done_length = 0

    def slice(self, request_range):
        """ 下载进度切片。"""
        put_begin, put_end = request_range
        if put_begin > self.begin + self.walk_length:
            if put_end != self.end:
                put_end = self.end
            if put_begin >= put_end:
                return None
        else:
            return None

        self._range = (self._range[0], put_begin)
        return put_begin, put_end

    def __repr__(self):
        return '<Progress [{}-{}]> {:.2%}'.format(self.begin, self.end, self.percent_complete / 100)

    def __iter__(self):
        return iter([self._range, self.walk_length, self.done_length])





from setuptools import setup, find_packages
import io
import os

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, 'nbdler', 'version.py'), 'r', encoding='utf-8') as f:
    exec(f.read(), about)

with io.open('README.rst', 'r', encoding='utf-8') as readme:
    long_description = readme.read()

install_requires = [
    'aiohttp',
    'requests'
]


setup(
    name=about['TITLE'],
    version=about['VERSION'],
    description=about['DESCRIPTION'],
    long_description=long_description,
    author=about['AUTHOR'],
    author_email=about['AUTHOR_EMAIL'],
    url=about['URL'],
    license=about['LICENSE'],
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
        ],
    packages=find_packages(),
    install_requires=install_requires,
)




import requests
from nbdler.uri import URIResponse
from traceback import format_exc
from .base_http import BaseHTTPClient, content_range_fullsize, content_type_mimetype
from nbdler.handler import h
import logging
import nbdler
from requests.utils import get_environ_proxies

log = logging.getLogger(__name__)


class HTTPClient(BaseHTTPClient):
    TIMEOUT = 10

    def connect(self):
        session = self.session
        source_uri = self.source_uri
        proxies = source_uri.proxies or {}
        if not proxies.get(source_uri.scheme):
            if source_uri.kwargs.get('trust_env', False):
                # Set environment's proxies.
                no_proxy = proxies.get('no_proxy') if proxies is not None else None
                env_proxies = get_environ_proxies(source_uri.uri, no_proxy=no_proxy)
                for (k, v) in env_proxies.items():
                    proxies.setdefault(k, v)

        cookies = source_uri.cookies
        verify = source_uri.kwargs.get('verify', True)
        uri, headers = self._build_uri_headers()
        timeout = self.kwargs.get('timeout', None) or HTTPClient.TIMEOUT
        try:
            resp = requests.get(
                source_uri.uri,
                headers=headers,
                proxies=proxies,
                cookies=cookies,
                timeout=timeout,
                stream=True,
                verify=verify
            )
        except requests.exceptions.Timeout as error:
            raise nbdler.error.TimeoutError(f"{uri}") from error
        except BaseException as error:
            log.debug(f'{error}', format_exc())
            raise nbdler.error.FatalError() from error
        else:
            total_length = content_range_fullsize(resp.headers.get('content-range'))
            response = URIResponse(str(resp.url), list(resp.headers.items()), resp.status_code, resp.reason,
                                   total_length, content_type_mimetype(resp.headers.get('content-type')),
                                   self.progress.range, resp.status_code == 206)

        if self.resume_capability is None:
            if resp.status_code not in (206, 200):
                raise nbdler.error.FatalError(f"[{resp.status_code} {resp.reason}] '{resp.url}'")
            self.resume_capability = resp.status_code == 206

        elif self.resume_capability is True:
            if not resp.status_code == 206:
                raise nbdler.error.FatalError(f"[{resp.status_code} {resp.reason}] '{resp.url}'")

        self.session = session
        self.resp = resp
        return response

    def fetch(self):
        session, resp = self.session, self.resp
        pg = self.progress

        speed_adjuster = h.speed_adjuster
        slicer = h.slicer
        uri_mgr = h.uri_mgr
        file_data = h.file_data
        receive_data = resp.raw.read

        pg.start()

        uri_mgr.success(resp)

        data = b''
        while True:
            if self._closed:
                break

            speed_adjuster.acquire_threadsafe()
            slicer.response_threadsafe()

            previous_len = len(data)
            remain_len = pg.total_length - pg.walk_length
            try:
                if remain_len >= 8192:
                    data += receive_data(8192)
                elif remain_len > 0:
                    data += receive_data(remain_len)
                else:
                    break
            except requests.exceptions.Timeout as err:
                uri_mgr.timeout(err)
                break
            except BaseException as err:
                uri_mgr.fatal(err)
                break

            walk_len = len(data) - previous_len
            if not walk_len:

                if resp.headers.get('content-length') is None:
                    pg.set_walk_finish()
                break

            pg.walk(walk_len)

            if pg.walk_length >= pg.total_length:
                break
            elif len(data) >= 65536:  # 64 KB
                file_data.store_threadsafe(data)
                data = b''
        if data:
            file_data.store_threadsafe(data)

        pg.stop()

    def run(self):
        h.slicer.response_threadsafe()
        try:
            resp = self.connect()
        except nbdler.error.UriError as err:
            h.uri_mgr.fatal(err)
            raise
        else:
            h.uri_mgr.success(resp)

            self.validate_token(resp)
            if not self._closed:
                self.fetch()

    @classmethod
    async def dlopen(cls, source, progress, **kwargs):
        with cls(None, source, progress, None, **kwargs) as cli:
            resp = cli.connect()
            return resp


def session_without_trust_env():
    session = requests.Session()
    # 默认创建不使用环境中的代理的会话，如要使用设置下载源的trust_env参数。
    session.trust_env = False
    return session


NAME = 'requests'
PROTOCOL_SUPPORT = ('http', 'https')
ASYNC_EXECUTE = False

ClientHandler = HTTPClient
ClientSession = session_without_trust_env






from . import aiohttp, requests
from .abstract import AbstractClient
from collections import defaultdict

__all__ = ['get_policy', 'ClientPolicy']

_solutions = defaultdict(list)
_name_solution = {}


class ClientPolicy:
    def __init__(self, **specified_mapping):
        self._specified = {k.lower(): v.lower() for k, v in specified_mapping.items()}

    def get_solution(self, protocol):
        """ 返回根据策略决定的客户端处理模块。
        Args:
            protocol: 要处理的协议

        Returns:
            返回客户端处理方案
        """
        sol_name = self._specified.get(protocol, None)
        if sol_name is None:
            # 使用该协议最新注册的客户端处理器作为默认的处理策略
            sol_name = _solutions.get(protocol, [None])[-1]
        if sol_name is None:
            raise NotImplementedError(f'没有找到协议{protocol}的处理策略。')
        solution = _name_solution.get(sol_name, None)
        if solution is None:
            raise NotImplementedError(f'没有找到名称为{sol_name}的客户端处理器。')
        return solution

    def __iter__(self):
        return iter(self._specified.items())


class ProtocolSolution:
    def __init__(self, module):
        self._module = module

    @property
    def name(self):
        return self._module.NAME

    @property
    def supported_protocols(self):
        return self._module.PROTOCOL_SUPPORT

    def is_async(self):
        return self._module.ASYNC_EXECUTE

    @property
    def dlopen(self):
        return self._module.ClientHandler.dlopen

    def get_client(self, *args, **kwargs):
        return self._module.ClientHandler(*args, **kwargs)

    def get_session(self, *args, **kwargs):
        return self._module.ClientSession(*args, **kwargs)


def get_policy(**kwargs):
    return ClientPolicy(**kwargs)


def register(module):
    """ 注册下载客户端处理模块。

    客户端模块规范：
    1. 客户端处理程序要求继承abstract_base.py中的AbstractClient类
    2. 使用类变量NAME作为客户端的唯一标识名称，尽量避免与其他客户端重名，
        重名的处理策略是后注册覆盖前注册。
    3. 使用ClientHandler作为客户端的类名，或通过赋值该模块变量名实现
    4. 使用ClientSession作为客户端会话，必须存在该变量，若不需要会话则赋值noop函数，
        客户端会话创建不提供参数，若需要提供使用functions.partial传递定义

    Args:
        module: 协议处理解决方案

    """
    global _solutions, _name_solution
    solution = ProtocolSolution(module)
    for protocol in solution.supported_protocols:
        _solutions[protocol].append(solution.name)
        _name_solution[solution.name] = solution


def main():
    # 多线程HTTP/HTTPS，使用requests库
    register(requests)
    # 异步HTTP/HTTPS，使用aiohttp库
    register(aiohttp)


# 注册下载客户端
main()



